'' +--------------------------------------------------------------------------+
'' | Cluso's RamBlade3 Driver                                          v0.203 |
'' +--------------------------------------------------------------------------+
'' |  Author:            "Cluso99" (Ray Rodrick)                              |
'' |  Copyright (c) 2009 "Cluso99" (Ray Rodrick)                              |
'' |  License            MIT License - See end of file for terms of use       |
'' +--------------------------------------------------------------------------+
                             
'' RR20091205   Derived from TriBladeProp_Blade2_203.spin for RamBlade
'' RR20091205   _203 add RamBlade
'' RR20091224      001
'' RR20091226      002
'' RR20110328      005 fix potential bug ROL/ROR -> SHL.SHR in rdloop/wrloop
'' +--------------------------------------------------------------------------+
'' RR20111018      006 version for RamBlade3

'' +--------------------------------------------------------------------------+
'' | Cluso's RamBlade3:  Hardware interface definitions                       |
'' +--------------------------------------------------------------------------+
'' NOTE: 1. This driver should be used to drive all hardware on the RamBlade except
''              the I/O on P30-31.
''       2. Before calling this routine, set DIRA=%xx000000_00000000_00000000_00000000 to avoid conflict.
''       3. P0...P29 are used by this routine, but P30 & P31 are not used.
''       6. Currently NO testing is done for ram address + length => 512KB (CAUSES BUS CONFLICT)
''
'' This driver may be extended to provide microSD, Flash and Eeprom routines.
''
' RamBlade3 hardware (see end of file for details)

VAR

' communication params(5) between cog driver code - only "command" and "errx" are modified by the driver
   long  command, hubaddrs, ramaddrs, blocklen, errx, cog ' rendezvous between spin and assembly (can be used cog to cog)
'        command  = R, W, S, M (read/write/set/move a block of sram) =0 when operation completed by cog
'        hubaddrs = hub address for data buffer
'        ramaddrs = ram address for data ($0_0000...$7_FFFF for 512KB total)
'        blocklen = ram buffer length for data transfer
'        errx     = returns =0 (false=good), else <>0 (true & error code)
'        cog      = cog no of driver (set by spin start routine)
   
PUB start : err_
' Initialise the RamBlade driver
  command := "I"
  cog := 1 + cognew(@rb_start, @command)
  if cog == 0
    err_ := $FF                 ' error = no cog
  else
    repeat while command        ' driver cog sets =0 when done
    err_ := errx                ' driver cog sets =0 if no error, else xx = error code

PUB stop
   if cog
      cogstop(cog~ - 1)      

PUB idle : err_
' Place the bus in the idle state (for SD driver use) [not reqd for RamBlade]
  err_ := 0

PUB active : err_
' Activate the bus (from SD driver use) [not reqd for RamBlade]
  err_ := 0


PUB DoCmd(command_, hub_address, ram_address, block_length) : err_
' Do the command: R, W, S, M (read/write/set/move a block of sram)
  hubaddrs := hub_address       ' hub address start
  ramaddrs := ram_address       ' ram address start
  blocklen := block_length      ' block length
  command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while command          ' driver cog sets =0 when done
  err_ := errx                  ' driver cog sets =0 if no error, else xx = error code


DAT
'' +--------------------------------------------------------------------------+
'' | Cluso's RamBlade3 Cog Driver                                             |
'' +--------------------------------------------------------------------------+
                        org     0
rb_start    ' setup the pointers to the hub command interface (saves execution time later)
                                      '  +-- These instructions are overwritten as variables after start
comptr        mov     comptr, par     ' -|  hub pointer to command                
hubptr        mov     hubptr, par     '  |  hub pointer to hub address            
ramptr        add     hubptr, #4      '  |  hub pointer to ram address            
lenptr        mov     ramptr, par     '  |  hub pointer to length                 
errptr        add     ramptr, #8      '  |  hub pointer to error status           
cmd           mov     lenptr, par     '  |  command  R/W/S/i/a                  
hubaddr       add     lenptr, #12     '  |  hub address                           
ramaddr       mov     errptr, par     '  |  ram address                           
len           add     errptr, #16     '  |  length                                
err           nop                     ' -+  error status returned (=0=false=good) 


' Initialise hardware [not reqd for RamBlade or RamBlade3???]
init                    mov     err, #0                 ' reset err=false=good
                        mov     dira, #0                ' all inputs
' set done & status
done                    wrlong  err, errptr             ' status  =0=false=good, else error x
                        wrlong  zero, comptr            ' command =0 (done)
' wait for a command
pause                   rdlong  cmd, comptr     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
' decode command
                        cmp     cmd, #"R"       wz      ' R = read block
              if_z      jmp     #rdblock
                        cmp     cmd, #"W"       wz      ' W = write block
              if_z      jmp     #wrblock
                        cmp     cmd, #"S"       wz      ' S = set (uses first hub byte as value to set)
              if_z      jmp     #setblock
'                       cmp     cmd, #"M"       wz      ' M = move (a ram block)
'             if_z      jmp     #???
                        mov     err, cmd                ' error = cmd (unknown command)
                        jmp     #done

'---------------------------------------------------------------------------------------------------------
'Memory Access Functions

'Read a "len" byte block given by "ramaddr" into hub at "hubaddr"
rdblock                 call    #ram_open               ' prepare for reading block
                                                        ' returns with outa = ramaddr, -CE=-WE=-OE=0
                        mov     dira, ram_dir_read      ' set outputs A0-18, CE, OE
                        nop                             ' sram delay required
rdloop                  mov     outx, ina               ' read byte from SRAM   \
                        shr     outx, #19               ' shift to lower 8 bits |
                       'and     outx,#$FF               ' remove other bits     | not reqd as wrbyte next 
                        wrbyte  outx, hubaddr           ' copy byte to hub      /
                        add     hubaddr, #1             ' inc hub pointer
                        add     outa, #1                ' inc sram address
                        djnz    len, #rdloop            ' loop for xxx bytes
                        jmp     #init

'Write a "len" byte block given by "ramaddr" from hub at "hubaddr"
'can optimise better
wrblock                 call    #ram_open               ' prepare for reading block
                                                        ' returns with outa = ramaddr, -CE=-WE=-OE=0
                        mov     dira, ram_dir_write     ' addr+data+CE outputs
wrloop                  rdbyte  outx, hubaddr           ' copy byte from hub    \
                        shl     outx, #19               ' place in bits 19-26   |
                        or      outx, ramaddr           ' add ramaddr bits      /
                        mov     outa, outx              ' address + data + -CE=-WE=-OE=0
                        mov     dira, ram_dir_wstrobe   ' addr+data+CE+WE outputs \ -WE=0 pulse
                        nop                             ' sram delay              |
                        mov     dira, ram_dir_write     ' addr+data+CE output     / end WE pulse
                        add     hubaddr, #1             ' inc hub pointer
                        add     ramaddr, #1             ' inc sram address
                        djnz    len, #wrloop            ' loop for xxx bytes
                        jmp     #init

'Set a "len" byte block given by "ramaddr" from byte at hub at "hubaddr"
'can optimise better
setblock                call    #ram_open               ' prepare for reading block
                                                        ' returns with outa = ramaddr, -CE=-WE=-OE=0
                        mov     dira, ram_dir_write     ' addr+data+CE outputs
                        rdbyte  tmpx, hubaddr           ' copy byte from hub    \
'                        rol     tmpx, #19               ' place in bits 19-26   /
                        shl     tmpx, #19               ' place in bits 19-26   / RR20111019 rol->shl better
setloop                 mov     outa, ramaddr           ' ramaddr bits          \
                        or      outa, tmpx              ' add byte to write     /
                        mov     dira, ram_dir_wstrobe   ' addr+data+CE+WE outputs \ -WE=0 pulse
                        nop                             ' sram delay              |
                        mov     dira, ram_dir_write     ' addr+data+CE output     / end WE pulse
                        add     ramaddr, #1             ' inc sram address
                        djnz    len, #setloop           ' loop for xxx bytes
                        jmp     #init


'SRAM access setup
ram_open                rdlong  hubaddr, hubptr         ' get hub address
                        rdlong  ramaddr, ramptr         ' get ram address
                        rdlong  len, lenptr             ' get length
                        mov     err, #5                 ' err=5
                        cmp     ramaddr, ram_top1 wc,wz ' < 512KB?
              if_ae     jmp     #done                   ' error "5"
                        mov     outa, ramaddr           ' setup read address incl -CE=-WE=-OE=0
ram_open_ret            ret



' RamBlade3 Hardware Access:
'
'       P0-18  = A0-18   SRAM address
'       P19-26 = D0-7    SRAM \ * see below
'                             | P26 = /OUT
'                             | P23 = /CS microSD (when P27=1)
'                             | P22 =  DI microSD
'                             | P21 = CLK microSD
'                             / P20 =  DO microSD
'       P27    = -CE     SRAM & enable microSD
'       P28    = -OE     SRAM & eeprom SCL
'       P29    = -WE     SRAM & eeprom SDA
'       P30    =  SO     serial/prog out
'       P31    =  SI     serial/prog in      
'
'                              +-------- OUT (P26) optional
'                              |  
'                  SRAM        |  +-------  -CS (microSD) (P23)
'                   -CE -----+ |  |+-------  DI (microSD) (P22)
'       (P28) SCL & -OE ----+| |  ||+------ CLK (microSD) (P21) 
'       (P29) SDA & -WE ---+|| |  |||+-----  DO (microSD) (P20)
'             (P30) SO ---+||| |  ||||    
'             (P31) SI --+|||| |  |||| 
'                        ||||| ||||||||
'          P31...P0 --> %00xxx_xxxxxxxx_xxxxxxxxxxxxxxxxxxx
'          10K pullups     ***    *
'                              +------+ +-----------------+
'                        xxWRE D7....D0 A18..............A0
ram_dir_read    long    %00011_00000000_1111111111111111111 ' outputs CE,     OE, A0-18, inputs D0-7, WE   (t/s)
ram_dir_write   long    %00001_11111111_1111111111111111111 ' outputs CE,         A0-18, D0-7, inputs WE=OE(t/s)
ram_dir_wstrobe long    %00101_11111111_1111111111111111111 ' outputs CE, WE,     A0-18, D0-7, inputs    OE(t/s)

zero          long      0              ' constant 0
outx          long      0              ' as a temp address + WE/OE/LE bits
tmpx          long      0              ' temp byte data
ram_top1      long      (512 * 1024)   ' 512KB (one sram)



{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |                                                            
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    | 
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}} 
